
require 'rexml/document'
require 'enumerator'

class Object
  def amrita_sanitize
    Amrita2::Sanitizer::sanitize_text(to_s)
  end

  def amrita_sanitize_as_attribute
    Amrita2::Sanitizer::sanitize_attribute_value(to_s)
  end

  def amrita_sanitize_as_url
    Amrita2::Sanitizer::sanitize_url(to_s)
  end
end

class String
  def amrita_sanitize
    Amrita2::Sanitizer::sanitize_text(self)
  end

  def amrita_sanitize_as_attribute
    Amrita2::Sanitizer::sanitize_attribute_value(self)
  end

  def amrita_sanitize_as_url
    Amrita2::Sanitizer::sanitize_url(self)
  end
end

module REXML
  class Child 
    def amrita_compile_main_code(cg)
      cg.puts(to_s)
    end
  end

  class Parent < Child
    def amrita_compile_main_code(cg)
      each_child do |c|
        c.amrita_compile_main_code(cg)
      end
    end
  end

  class Element < Parent
    attr_accessor :amrita_spec
    def amrita_compile_main_code(cg)
      cg.compile_element_main_code(self) do
        super
      end
    end

    def amrita_compile_sub_code(cg)
      cg.compile_element_sub_code(self) do
        elements.each do |e|
          e.amrita_compile_sub_code(cg)
        end
      end
    end
  end
end

module Amrita2
  # this is a String that amrita dosen't sanitize.
  # If you don't sanitize or sanitize yourself,
  # pass SanitizedString.new(x) as model data.
  class SanitizedString < String
    def SanitizedString::[](s)
      new(s).freeze
    end

    def amrita_sanitize
      self
    end

    def amrita_sanitize_as_attribute
      self
    end

    def amrita_sanitize_as_url
      self
    end

    def to_s
      self
    end
  end

  # This module provide methods for avoid XSS vulnerability
  # taken from IPA home page(Japanese)
  # http://www.ipa.go.jp/security/awareness/vendor/programming/a01_02.html
  module Sanitizer
    NAMECHAR = '[-\w\d\.:]'
    NAME = "([\\w:]#{NAMECHAR}*)"
    NOT_REFERENCE = "(?!#{NAME};|&#\\d+;|&#x[0-9a-fA-F]+;)" # borrowed from rexml
    AMP_WITHOUT_REFRENCE = /&#{NOT_REFERENCE}/
    # escape &<>
    def sanitize_text(text)
      s = text.dup
      s.gsub!(AMP_WITHOUT_REFRENCE, '&amp;')
      s.gsub!("<", '&lt;')
      s.gsub!(">", '&gt;')
      s
    end

    # escape &<>"'
    def sanitize_attribute_value(text)
      s = text.dup
      s.gsub!(AMP_WITHOUT_REFRENCE, '&amp;')
      s.gsub!("<", '&lt;')
      s.gsub!(">", '&gt;')
      s.gsub!('"', '&quot;')
      s.gsub!("'", '&#39;')
      s
    end

    DefaultAllowedScheme = {
      'http' => true,
      'https' => true,
      'ftp' => true,
      'mailto' => true,
    }

    #UrlInvalidChar = Regexp.new(%q|[^;/?:@&=+$,A-Za-z0-9\-_.!~*'()%]|)
    UrlInvalidChar = Regexp.new(%q|[^;/?:@&=+$,A-Za-z0-9\-_.!~*'()%#]|) #'

    # +sanitize_url+ accepts only these characters 
    #     --- http://www.ietf.org/rfc/rfc2396.txt ---
    #     uric = reserved | unreserved | escaped
    #     reserved = ";" | "/" | "?" | ":" | "@" | "&" | "=" | "+" | "$" | ","
    #     unreserved = alphanum | mark
    #     mark = "-" | "_" | "." | "!" | "~" | "*" | "'" | "(" | ")"
    #     escaped = "%" hex hex
    # 
    # +sanitize_url+ accepts only schems specified by +allowd_scheme+
    #
    # The default is http: https: ftp: mailt:

    def  sanitize_url(text, allowd_scheme = DefaultAllowedScheme)
      # return nil if text has characters not allowd for URL

      return nil if text =~ UrlInvalidChar

      # return '' if text has an unknown scheme
      # --- http://www.ietf.org/rfc/rfc2396.txt ---
      # scheme = alpha *( alpha | digit | "+" | "-" | "." )

      if text =~ %r|^([A-Za-z][A-Za-z0-9+\-.]*):| 
        return nil unless allowd_scheme[$1]
      end
  
      # escape HTML
      # special = "&" | "<" | ">" | '"' | "'"
      # But I checked  "<" | ">" | '"' before.
      s = text.dup
      #s.gsub!("&", '&amp;')
      s.gsub!("'", '&#39;')

      s
    end

    module_function :sanitize_text, :sanitize_attribute_value, :sanitize_url
  end

  module Core
    class CodeGenerator
      attr_accessor :current_stream

      def initialize(stream = "get_mainstream")
        @current_stream = stream
        init_results
        init_strbuf
        @current_indent = 0
        @dynamic_element_stack = []
      end

      def puts(msg)
        @strbuf.concat msg
      end

      def flush
        if @strbuf.size > 0
          @results << [@current_indent, "#{current_stream} << " + @strbuf.inspect]
          init_strbuf
        end
      end

      def code(l)
        flush
        @results << [@current_indent, l]
      end

      def results
        flush
        @results.collect do |level, line|
          "  " * level + line
        end
      end

      def if_(cond)
        code("if #{cond}")
        level_up do
          yield
        end
        code("end")
      end

      def else_
        level_down do
          code("else")
        end
        yield
      end

      def level_up
        flush
        @current_indent += 1
        yield
      ensure
        flush
        @current_indent -= 1
      end

      def level_down
        flush
        @current_indent -= 1
        yield
      ensure
        flush
        @current_indent += 1
      end

      def define_method(name, *params)
        if params.size == 0
          code("def #{name}")
        else
          code("def #{name}(#{params.join(',')})")
        end
        level_up do
          yield
        end
        code("end")
      end

      def define_module(name)
        code("module #{name}")
        level_up do
          yield
        end
        code("end")
      end

      def compile_element_main_code(element, &block)
        spec = element.amrita_spec
        if spec
          spec.compile_element_main_code(element, self, &block)
        else
          generate_static_element_code(element, &block)
        end
      end

      def compile_element_sub_code(element, &block)
        spec = element.amrita_spec
        if spec
          spec.compile_element_sub_code(element, self, &block) 
        else
          yield
        end
      end

      def init_results
        @results = []
      end

      def init_strbuf
        @strbuf = ""
      end

      def generate_static_element_code(e, option={}, &block)
        if e.name and e.name.size > 0
          unless option[:delete_tag]
            a = []
            a << e.name 
            if (option[:delete_id])
              a += e.attributes.enum_for(:each_attribute).find_all do |attr|
                attr.name != "id"
              end.collect do |attr| 
                attr.to_string 
              end
            else
              a += e.attributes.enum_for(:each_attribute).collect { |attr| attr.to_string }
            end
            puts("<#{a.join(' ')}>")
          end
          yield
          puts("</#{e.name}>") unless option[:delete_tag]
        else
          yield
        end
      end

      def new_context(new_stream,  &block)
        code("new_context(#{new_stream}) do")
        level_up do
          yield
        end
        code("end")
      end

      def dynamic_element_push(de)
        @dynamic_element_stack.unshift de
      end

      def dynamic_element_top
        @dynamic_element_stack.first
      end

      def dynamic_element_pop
        @dynamic_element_stack.shift
      end
    end

    class Matcher
      def find(e)
        return e if match(e)
        e.elements.each do |c|
          x = find(c)
          return x if x
        end
        nil
      end
    end

    class AttrMatcher < Matcher
      attr_reader :sym
      def initialize(sym, val)
        @sym = sym
        @sym_s = sym.to_s
        @val = val
      end

      def match(e)
        if @val != '*'
          @val == e.attributes[@sym_s]
        else
          e.attributes[@sym_s]
        end
      end
    end

    class IDMatcher < AttrMatcher
      def initialize(val)
        super(:id, val.to_s)
      end
    end

    class TagMatcher < Matcher
      def initialize(tag)
        @tag = tag.to_s
      end

      def match(e)
        e.name == @tag
      end
    end

    class AnyMatcher < Matcher
      def match(e)
        true
      end
    end

    module Runtime
      def new_element(tag, attrs={})
        ret = REXML::Element.new(tag.to_s)
        attrs.each do |k,v|
          ret.attributes[k.to_s] = v.to_s
        end
        ret
      end
      
      def start_tag(e, out="")
        out << "<#{e.expanded_name}"
        e.attributes.each_attribute do |attr|
          next if attr.value == ""
          out << " "
          attr.write(out)  
        end unless e.attributes.empty?
        out << ">"
        out
      end

      def end_tag(e, out="")
        out << "</#{e.expanded_name}>"
        out
      end

      class Context 
        attr_reader :main_stream
        
        def initialize(stream)
          @main_stream = stream
          @sub_stream = {}
        end

        def sub_stream(key)
          @sub_stream[key] ||= ""
        end

        def clear_substream(key)
          @sub_stream[key] = ""
        end
      end
      
      def context_stack
        Thread::current[:amrita_context_stack] ||= []
      end

      def current_context
        context_stack.first
      end

      def new_context(stream="", &block)
        context_stack.unshift Context.new(stream)
        yield
        if stream.kind_of?(String)
          SanitizedString[stream]
        else
          stream
        end
      ensure
        context_stack.shift
      end

      def get_mainstream
        current_context.main_stream
      end

      def get_substream(key)
        current_context.sub_stream(key)
      end

      def output_substream(key)
        c = current_context
        s = c.sub_stream(key)
        c.main_stream << SanitizedString[s]
        c.clear_substream(key)
      end
    end
    

    class DynamicElementSpec
      attr_accessor :cloned
      attr_reader :sym, :matcher, :children, :option, :assigned_elements

      def initialize(sym, option={}, &block)
        case sym
        when Symbol
          @sym = sym
        when nil
          @sym = nil
        else
          @sym = sym.to_s.intern
        end
        @cloned = false
        @option, @children = option, []
        @direct_define = @option[:direct_define]
        setup_matcher
        setup_generator
        @in_setup = true
        self.instance_eval(&block) if block_given?
        @in_setup = false
      end

      def clone
        new_spec = DynamicElementSpec.new(sym, option)
        @children.each_with_index do |child, i|
          new_spec.children[i] = child.clone
        end
        new_spec.cloned = true
        new_spec
      end

      def assign_element(e)
        raise "can't assign not cloned dynamic spec" unless @cloned
        @assigned_elements ||= []
        @assigned_elements << e
      end

      def matched_many?
        @option[:match_many] and @assigned_elements.size > 1 
      end

      def first_match?(element)
        matched_many? and @assigned_elements.index(element) == 0
      end

      def name_for(element)
        if matched_many?
          i = @assigned_elements.index(element)
          "#{@sym}_#{i}".intern
        else
          @sym
        end
      end

      def dynamic_element(name, option={}, &block)
        option[:direct_define] = @direct_define
        @children << DynamicElementSpec.new(name, option, &block)
      end

      
      def compile_element_main_code(element, cg, &block)
        @generator.generate_main(element, self, cg, &block)
      end

      def compile_element_sub_code(element, cg, &block)
        @generator.generate_sub(element, self, cg, &block)
      end

      def method_missing(sym, params={}, &block)
        if @direct_define and @in_setup
          dynamic_element(sym, params, &block)
        else
          super
        end
      end

      def module_name
        sym.to_s.capitalize  
      end

      def output_stream
        @option[:output_stream] or sym
      end

      private
      def setup_matcher
        if sym
          @matcher = @option[:matcher] || IDMatcher.new(sym)
        else
          @matcher = AnyMatcher.new
        end
      end

      def setup_generator
        if @option[:generator] 
          @generator = @option[:generator]
        else
          @generator = DefaultCodeGenerator.new
          @generator.extend CheckTrue if @option[:check_true]
          @generator.extend ReplaceAttr if @option[:replace_attr]
          @generator.extend UseOriginalElement if @option[:use_original_element]
          @generator.extend ExpandByMember if @option[:expand_by_member]
        end
      end

      def define_expand_template(cg, e)
        cg.code("new_context(out) do")
        cg.level_up do
          cg.code("block.call(#{module_name or 'nil'}) if block_given? ")
          e.amrita_compile_main_code(cg)
        end
        cg.code("end")
      end

      def define_sub_modules(cg, e)
        e.amrita_compile_sub_code(cg)
      end
    end

    class RootSpec < DynamicElementSpec
      def initialize(option={}, &block)
        super(nil, option, &block)
      end

      def module_name
        "MainContext"
      end

      def create_template(e)
        case e
        when REXML::Document
          root = e.root
        else
          root = REXML::Document.new(e).root
        end

        ret = Object.new
        m = create_template_module(root)
        ret.extend(m)
        ret
      end

      def create_template_module(e)
        m = Module.new
        result = create_template_module_text(e)
        puts result  if $DEBUG
        m.module_eval result
        m
      end

      def create_template_module_text(e)
        specs = [self.clone]
        traverse_and_search_match(specs, e)
        # not match check
        raise "SPEC #{specs.first.sym} was not matched" if specs.size > 0 

        cg = CodeGenerator.new
        cg.code("include Amrita2::Runtime")

        cg.define_method(:expand_template, :out, "&block") do
          define_expand_template(cg, e)
        end

        cg.define_module(module_name) do
          cg.code("extend Amrita2::Runtime")
          define_sub_modules(cg, e)
        end
        cg.results.join("\n")
      end

      def traverse_and_search_match(child_specs, element)
        s = child_specs.first
        return unless s
        if s.matcher.match(element)
          element.amrita_spec = s
          s.assign_element(element)
          child_specs.shift unless s.option[:match_many]
          specs = s.children.clone
          traverse_and_search_match(specs, element)

          # not match check
          raise "SPEC #{specs.first.sym} was not matched" if specs.size > 0 and ! specs.first.option[:match_many]

          return true
        else
          element.elements.each do |ce|
            ret = traverse_and_search_match(child_specs, ce)
          end
          return false
        end
      end

    end

    class DefaultCodeGenerator
      def generate_main(element, dyn_spec, cg, &block)
        sym = element.amrita_spec.sym
        if sym
          generate_main_code(element, dyn_spec, cg, sym.to_s, &block)
        else
          cg.generate_static_element_code(element, &block)
        end
      end

      def generate_sub(element, dyn_spec, cg, &block)
        generate_dynamic_element_code(element, dyn_spec, cg, &block)
      end

      private
      def generate_main_code(element, dyn_spec, cg, sym)
        cg.code("output_substream(:#{sym})")
      end

      def generate_dynamic_element_code(element, dyn_spec, cg, &block)
        sym = dyn_spec.sym
        if dyn_spec.children.size == 0 
          code_for_leaf_element(element, dyn_spec, cg, &block)
        else
          if sym
            code_for_not_leaf_element(element, dyn_spec, cg, &block)
          else
            yield
          end
        end
      end

      def code_for_leaf_element(element, dyn_spec, cg, &block)
        sym = dyn_spec.sym
        cg.define_method(sym, "val", "attrs={}") do
          cg.new_context("get_substream(:#{dyn_spec.output_stream})") do
            code_for_element_and_scalar(element, dyn_spec, cg)
          end
        end
        cg.code("module_function :#{sym}")
      end

      def code_for_not_leaf_element(element, dyn_spec, cg, &block)
        if dyn_spec.matched_many? and dyn_spec.first_match?(element)
          make_selection_code(element, dyn_spec, cg, &block)
        end
        sym = dyn_spec.name_for(element)
        cg.define_method(sym, "attrs={}", "&block") do
          cg.new_context("get_substream(:#{dyn_spec.output_stream})") do
            code_for_element_with_child(element, dyn_spec, cg, &block)
          end
        end
        cg.code("module_function :#{sym}")
        if ! dyn_spec.matched_many? or dyn_spec.first_match?(element)
          cg.define_module(dyn_spec.module_name) do
            cg.code("extend Amrita2::Runtime")
            yield
          end
        end
      end

      def make_selection_code(element, dyn_spec, cg, &block)
        sym = dyn_spec.sym
        cg.define_method(sym, "no=0", "attrs={}", "&block") do
          cg.code("no = no % #{dyn_spec.assigned_elements.size}")
          cg.code("mname = (#{sym.to_s.inspect} + '_' + no.to_s).intern")
          cg.code("self.send(mname, attrs, &block)")
        end        
        cg.code("module_function :#{sym}")
      end

      def code_for_element_and_scalar(element, dyn_spec, cg)
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.code("get_mainstream << val.amrita_sanitize")
        end
      end

      def code_for_element_with_child(element, dyn_spec, cg)
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.code("yield(#{dyn_spec.module_name})")
          element.each_child do |c|
            c.amrita_compile_main_code(cg)
          end
          #element.elements.each do |e|
          #  e.amrita_compile_main_code(cg)
          #end
        end
      end

      def gen_element_with_attr(element,cg, varname='e')
        a = element.attributes.collect do |k, v|
          ":#{k} => #{v.inspect}, "
        end
        default_value = "{ #{a.join(' ')}}"
        cg.code("a = #{default_value}")
        cg.code("attrs.each do |k,v|")
        cg.level_up do
          cg.code("a[k] = v")
        end
        cg.code("end")
        cg.code("#{varname} = new_element(#{element.name.inspect}, a)")
      end
    end
  end

  module CheckTrue
    def code_for_leaf_element(element, dyn_spec, cg, &block)
      sym = dyn_spec.sym
      cg.define_method(sym, :val) do
        cg.new_context("get_substream(:#{dyn_spec.output_stream})") do
          cg.if_("val == true") do
            cg.puts(element.to_s)
            cg.else_ do
              code_for_element_and_scalar(element, dyn_spec, cg)
            end
          end
        end
      end
      cg.code("module_function :#{sym}")
    end
  end

  module ReplaceAttr
    def code_for_element_and_scalar(element, dyn_spec, cg)
      cg.if_("attrs.size == 0") do
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.code("get_mainstream << val.amrita_sanitize")
        end
        cg.else_ do
          gen_element_with_attr(element, cg)
          cg.code("get_mainstream << start_tag(e) << val.amrita_sanitize << end_tag(e)")
        end
      end
    end

    def code_for_element_with_child(element, dyn_spec, cg)
      cg.if_("attrs.size == 0") do
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.code("yield(#{dyn_spec.module_name})")
          element.elements.each do |e|
            e.amrita_compile_main_code(cg)
          end
        end
        cg.else_ do
          gen_element_with_attr(element, cg)
          cg.code("get_mainstream << start_tag(e)")
          cg.code("yield(#{dyn_spec.module_name})")
          element.each_child do |e|
            e.amrita_compile_main_code(cg)
          end
          cg.code("get_mainstream << end_tag(e)")
        end
      end
    end

  end

  module UseOriginalElement
    def code_for_leaf_element(element, dyn_spec, cg, &block)
      sym = dyn_spec.sym
      cg.define_method(sym, "val=nil", "attrs={}") do
        cg.new_context("get_substream(:#{dyn_spec.output_stream})") do
          cg.if_("block_given?") do
            gen_element_with_attr(element, cg)
            cg.code("e.text = #{element.text.inspect}")
            cg.code("get_mainstream << yield(e).to_s")
            cg.else_ do
              code_for_element_and_scalar(element, dyn_spec, cg)
            end
          end
        end
      end
      cg.code("module_function :#{sym}")
    end

    def code_for_element_with_child(element, dyn_spec, cg)
      cg.code("yield(#{dyn_spec.module_name})")
      cg.if_("get_substream(:#{dyn_spec.output_stream}).size == 0") do
        cg.generate_static_element_code(element, dyn_spec.option) do 
          element.each_child do |e|
            e.amrita_compile_main_code(cg)
          end
        end
        cg.else_ do
          cg.code("output_substream(:#{dyn_spec.name_for(element)})")
          element.each_child do |e|
            e.amrita_compile_main_code(cg)
          end
          cg.puts(end_tag(element))
        end
      end
    end

    def code_for_not_leaf_element(element, dyn_spec, cg, &block)
      if dyn_spec.matched_many? and dyn_spec.first_match?(element)
        make_selection_code(element, dyn_spec, cg, &block)
      end
      sym = dyn_spec.name_for(element)
      cg.define_method(sym, "attrs={}", "&block") do
        cg.new_context("get_substream(:#{sym})") do
          code_for_element_with_child(element, dyn_spec, cg, &block)
        end
      end
      cg.code("module_function :#{sym}")

      if ! dyn_spec.matched_many? or dyn_spec.first_match?(element)
        cg.define_module(dyn_spec.module_name) do
          cg.code("extend Amrita2::Core::Runtime")
          cg.define_method(sym, "val=nil", "attrs={}") do
            cg.new_context("get_substream(:#{sym})") do
              cg.code("e = yield(new_element(#{element.name.inspect}))") 
              cg.code("get_mainstream << start_tag(e)")
            end
            yield
          end
          cg.code("module_function :#{sym}")
        end
      end
    end
  end

  module ExpandByMember
    def code_for_not_leaf_element(element, dyn_spec, cg, &block)
      if dyn_spec.matched_many? and dyn_spec.first_match?(element)
        make_selection_code(element, dyn_spec, cg, &block)
      end
      sym = dyn_spec.name_for(element)
      if ! dyn_spec.matched_many? or dyn_spec.first_match?(element)
        cg.define_method(sym, "val=nil", "attrs={}", "&block") do
          cg.new_context("get_substream(:#{dyn_spec.output_stream})") do
            code_for_element_with_child(element, dyn_spec, cg, &block)
          end
        end
        cg.code("module_function :#{sym}")
        cg.define_module(dyn_spec.module_name) do
          cg.code("extend Amrita2::Runtime")
          yield
        end
      end
    end

    def code_for_element_with_child(element, dyn_spec, cg)
      cg.generate_static_element_code(element, dyn_spec.option) do 
        cg.if_("block_given?") do
          cg.code("yield(#{dyn_spec.module_name})")
          cg.else_ do
            dyn_spec.children.each do |c|
              cg.code("#{dyn_spec.module_name}::#{c.sym}(val.#{c.sym})")
            end
          end
        end
        element.each_child do |e|
          e.amrita_compile_main_code(cg)
        end
      end
    end
  end

  def self.make_default_spec(root, option={:delete_id=>true, :check_true=>true})
    s = Core::RootSpec.new(option.clone)
    make_default_spec1(root, option, [s])
    s
  end

  def self.make_default_spec1(e, option, spec_stack)
    id_ = e.attributes["id"]
    if id_
      sym = id_.intern
      s = Core::DynamicElementSpec.new(sym, option.clone)
      spec_stack.first.children << s
      spec_stack.unshift(s)
      e.elements.each do |c|
        make_default_spec1(c, option, spec_stack)
      end
      spec_stack.shift
    else
      e.elements.each do |c|
        make_default_spec1(c, option, spec_stack)
      end
    end
  end

  def self.default_expand(t, out, data)
    t.expand_template(out) do |m|
      default_expand1(m, data)
    end
  end

  def self.default_expand1(m, data)
    data.each do |k, v|
      case v
      when String, true, false, nil
        m.send(k, v)
      when Hash
        m.send(k) do |m2|
          default_expand1(m2, v)
        end
      when Enumerable
        v.each do |x|
          case x
          when Hash
            m.send(k) do |m2|
              default_expand1(m2, x)
            end
          else
            m.send(k, x)
          end
        end
      else
        m.send(k, v.to_s)
      end
    end
  end

  module HtmlTag
    include Amrita2::Core
    HEAD = TagMatcher.new(:head)
    BODY = TagMatcher.new(:BODY)
    TABLE = TagMatcher.new(:table)
    TR = TagMatcher.new(:tr)
    TH = TagMatcher.new(:th)
    TD = TagMatcher.new(:td)
    UL = TagMatcher.new(:ul)
    OL = TagMatcher.new(:ol)
    LI = TagMatcher.new(:li)
    A = TagMatcher.new(:a)
  end

  Runtime = Core::Runtime
  def self.define_tempalte_spec(option={}, &block)
    Core::RootSpec.new(option, &block)
  end
end

